@@ -6,11 +6,12 @@ use v5.10;
 use parent 'Plack::Component';
 use Plack::Util::Accessor qw(hook access app events safe);
 use Plack::Request;
+use Plack::Builder;
 use Plack::Middleware::Access;
 use Carp qw(croak);
 use JSON qw(decode_json);
 
-our $VERSION = '0.6';
+our $VERSION = '0.7';
 
 sub prepare_app {
     my $self = shift;
@@ -31,12 +32,11 @@ sub prepare_app {
         deny  => "all"
     ]) unless $self->access;
 
-    $self->app(
-        Plack::Middleware::Access->wrap(
-            sub { $self->call_granted($_[0]) },
-            rules => $self->access
-        )
-    );
+    $self->app( builder {
+        enable 'Access', rules => $self->access;
+        enable 'HTTPExceptions';
+        sub { $self->call_granted($_[0]) },
+    } );
 }
 
 sub call {
@@ -54,29 +54,22 @@ sub call_granted {
     my $req = Plack::Request->new($env);
     my $event = $env->{'HTTP_X_GITHUB_EVENT'} // '';
     my $delivery = $env->{'HTTP_X_GITHUB_DELIVERY'} // '';
-    my $json;
+    my $payload;
+    my ($status, $message);
     
     if ( !$self->events or grep { $event eq $_ } @{$self->events} ) {
-        $json = eval { decode_json $req->content };
+        $payload = eval { decode_json $req->content };
     }
 
-    if (!$json) {
+    if (!$payload) {
         return [400,['Content-Type'=>'text/plain','Content-Length'=>11],['Bad Request']];
     }
     
-    my $log = sub {
-        $env->{'psgix.logger'}->({ level => $_[0], message => $_[1] });
-    };
-    my $logger = Plack::Util::inline_object(
-        debug => sub { $log->( debug => $_[0] ) },
-        info  => sub { $log->( info  => $_[0] ) },
-        warn  => sub { $log->( warn  => $_[0] ) },
-        error => sub { $log->( error => $_[0] ) },
-        fatal => sub { $log->( fatal => $_[0] ) },
+    my $logger = Plack::App::GitHub::WebHook::Logger->new(
+        $env->{'psgix.logger'} || sub { }
     );
 
-    my ($status, $message);
-    if ( $self->receive($json, $event, $delivery, $logger) ) {
+    if ( $self->receive( [ $payload, $event, $delivery, $logger ], $env->{'psgi.errors'} ) ) {
         ($status, $message) = (200,"OK");
     } else {
         ($status, $message) = (202,"Accepted");
@@ -84,25 +77,61 @@ sub call_granted {
 
     $message = ucfirst($event)." $message" if $self->events;
 
-    [ $status,
+    return [ 
+        $status,
         [ 'Content-Type' => 'text/plain', 'Content-Length' => length $message ],
-        [$message] ];
+        [ $message ] 
+    ];
 }
 
 sub receive {
-    my $self = shift;
+    my ($self, $args, $error) = @_;
 
     foreach my $hook (@{$self->{hook}}) {
-        if ($self->safe) {
-            return unless eval { $hook->(@_) } and !$@;
-        } else {
-            return unless $hook->(@_);
+        if ( !eval { $hook->(@$args) } || $@ ) {
+            if ( $@ ) {
+                if ($self->safe) {
+                    $error->print($@);
+                } else {
+                    die Plack::App::GitHub::WebHook::Exception->new( 500, $@ );
+                }
+            }
+            return;
         }
     } 
 
     return scalar @{$self->{hook}};
 }
 
+{
+    package Plack::App::GitHub::WebHook::Logger;
+    sub new {
+        my $self = bless { logger => $_[1] }, $_[0];
+        foreach my $level (qw(debug info warn error fatal)) {
+            $self->{$level} = sub { $self->log( $level => $_[0] ) }
+        }
+        $self;
+    }
+    sub log {
+        my ($self, $level, $message) = @_;
+        chomp $message;
+        $self->{logger}->({ level => $level, message => $message });
+        1;
+    }
+    sub debug { $_[0]->log(debug => $_[1]) }
+    sub info  { $_[0]->log(info  => $_[1]) }
+    sub warn  { $_[0]->log(warn  => $_[1]) }
+    sub error { $_[0]->log(error => $_[1]) }
+    sub fatal { $_[0]->log(fatal => $_[1]) }
+}
+
+{
+    package Plack::App::GitHub::WebHook::Exception;
+    use overload  '""' => sub { $_[0]->{message} };
+    sub new { bless { code => $_[1], message => $_[2] }, $_[0]; }
+    sub code { $_[0]->{code} }
+}
+
 1;
 __END__
 
@@ -128,7 +157,7 @@ Plack::App::GitHub::WebHook - GitHub WebHook receiver as Plack application
 
     Plack::App::GitHub::WebHook->new(
         hook => sub {
-            my $payload = shift;
+            my ($payload, $event, $delivery, $logger) = @_;
             ...
         }
     )->to_app;
@@ -144,9 +173,11 @@ are called one by one until a task returns a false value.
     Plack::App::GitHub::WebHook->new(
         hook => [
             sub { $_[0]->{repository}{name} eq 'foo' }, # filter
-            { Filter => { repository_name => 'foo' } }, # equivalent filter
-            sub { my ($payload) = @_; ...  }, # some action
-            sub { run3 \@cmd ... }, # some more action
+            sub { # action
+                my ($payload, $event, $delivery, $logger) = @_;
+                run3 \@cmd, undef, $logger->{info}, $logger->{error}; 
+            },
+            sub { ...  }, # some more action
         ]
     )->to_app;
 
@@ -183,30 +214,33 @@ By default access is restricted to known GitHub WebHook IPs.
 The following application automatically pulls the master branch of a GitHub
 repository into a local working directory.
 
-    use Git::Repository;
     use Plack::App::GitHub::WebHook;
+    use IPC::Run3;
 
     my $branch = "master;
     my $work_tree = "/some/path";
 
     Plack::App::GitHub::WebHook->new(
         events => ['push','ping'],
-        safe => 1,
         hook => [
             sub { 
-                my ($payload, $method) = @_;
-                $method eq 'ping' or $payload->{ref} eq "refs/heads/$branch";
+                my ($payload, $event, $delivery, $log) = @_;
+                $log->info("$event $delivery");
+                $event eq 'ping' or $payload->{ref} eq "refs/heads/$branch";
             },
             sub {
-                my ($payload, $method) = @_;
-                return 1 if $method eq 'ping'; 
+                my ($payload, $event, $delivery, $log) = @_;
+                my $origin = $payload->{repository}->{clone_url} 
+                           or die "missing clone_url\n";
+                my $cmd;
                 if ( -d "$work_tree/.git") {
-                    Git::Repository->new( work_tree => $work_tree )
-                                   ->run( 'pull', origin => $branch );
+                    chdir $work_tree;
+                    $cmd = ['git','pull',$origin,$branch];
                 } else {
-                    my $origin = $payload->{repository}->{clone_url};
-                    Git::Repository->run( clone => $origin, -b => $branch, $work_tree );
+                    $cmd = ['git','clone',$origin,'-b',$branch,$work_tree];
                 }
+                $log->info(join ' ', '$', @$cmd);
+                run3 $cmd, undef, $log->{info}, $log->{error};
                 1;
             },
             # sub { ...optional action after each pull... } 
@@ -245,6 +279,10 @@ Otherwise, if the hook was called and returned a true value.
 
 Otherwise, if the hook was called and returned a false value.
 
+=item HTTP 500 Internal Server Error
+
+If a hook died with an exception, the error is returned as content body.                
+
 =back
 
 This module requires at least Perl 5.10.
@@ -258,9 +296,9 @@ This module requires at least Perl 5.10.
 A code reference or an array of code references with tasks that are executed on
 an incoming webhook.  Each task gets passed the encoded payload, the
 L<event|https://developer.github.com/webhooks/#events> and the unique delivery
-ID.  If the task returns a true value, next the task is called or HTTP status
-code 200 is returned.  Information can be passed from one task to the next by
-modifying the payload. 
+ID, and a L<logger object|/LOGGING>.  If the task returns a true value, next
+the task is called or HTTP status code 200 is returned.  Information can be
+passed from one task to the next by modifying the payload. 
 
 If a task returns a false value or if no task was given, HTTP status code 202
 is returned immediately. This mechanism can be used for conditional hooks or to
@@ -269,8 +307,22 @@ reason.
 
 =item safe
 
-Wrap all hook tasks in C<< eval { ... } >> blocks to catch exceptions. A dying
-task in safe mode is equivalent to a task that returns a false value.
+Wrap all hook tasks in C<< eval { ... } >> blocks to catch exceptions.  Error
+messages are send to the PSGI error stream C<psgi.errors>.  A dying task in
+safe mode is equivalent to a task that returns a false value, so it will result
+in a HTTP 202 response.
+
+Plack::Middleware::HTTPExceptions
+
+If you want errors to result in a HTTP 500 response,
+wrap the application in an eval block such as this:
+
+    sub {
+        eval { $app->(@_) } || do {
+            my $msg = $@ || 'Server Error';
+            [ 500, [ 'Content-Length' => length $msg ], [ $msg ] ];
+        };
+    };
 
 =item access
 
@@ -288,6 +340,29 @@ expected to be send with the C<X-GitHub-Event> header (e.g. C<['pull']>).
 
 =back
 
+=head1 LOGGING
+
+Each hook is passed a logging object as fourth parameter. It provides logging
+methods for each log level and a general log method:
+
+    sub sample_hook {
+        my ($payload, $event, $delivery, $log) = @_;
+
+        $log->debug('message');  $log->{debug}->('message');
+        $log->info('message');   $log->{info}->('message');
+        $log->warn('message');   $log->{warn}->('message');
+        $log->error('message');  $log->{error}->('message');
+        $log->fatal('message');  $log->{fatal}->('message');
+
+        $log->log( warn => 'message' );
+
+        run3 \@system_command, undef,
+            $log->{info},   # STDOUT to log level info
+            $log->{error};  # STDERR to log level error
+    }
+
+Trailing newlines on log messages are trimmed.
+
 =head1 DEPLOYMENT
 
 Many deployment methods exist. An easy option might be to use Apache webserver
@@ -299,7 +374,7 @@ Plack::App::GitHub::WebHook:
     sudo cpanm Plack::App::GitHub::WebHook
 
 Then add this section to C</etc/apache2/sites-enabled/default> (or another host
-configuration) and restart apache afterwards (C<sudo service apache2 restart>):
+configuration) and restart Apache.
 
     <Directory /var/www/webhooks>
        Options +ExecCGI -Indexes +SymLinksIfOwnerMatch
